# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import string, keyword
from gamera import util

class SymbolTable(util.CallbackObject):
   def __init__(self):
      util.CallbackObject.__init__(self)
      self.symbols = {}
      self.add("skip", 0)

   ########################################
   # CALLBACKS

   def add_callback(self, alert, callback):
      util.CallbackObject.add_callback(self, alert, callback)
      if alert == "add":
         for symbol in self.symbols.keys():
            callback(self.normalize_symbol(symbol)[1])

   ########################################

   def normalize_symbol(self, symbol):
      # assert type(symbol) == types.StringType
      while len(symbol) and symbol[0] == '.':
         symbol = symbol[1:]
      if symbol == '':
         return '', []
      symbol = symbol.lower()
      orig_tokens = symbol.strip().split('.')
      tokens = []
      for token in orig_tokens:
         if token.strip() != '':
            tokens.append(token.strip())
      symbol = '.'.join(tokens)
      return symbol, tokens

   def add(self, symbol, id = -1):
      symbol, tokens = self.normalize_symbol(symbol)
      if self.symbols.has_key(symbol):
         return symbol
      self.symbols[symbol] = None
      self.trigger_callback('add', tokens)
      return symbol

   def remove(self, symbol):
      symbol, tokens = self.normalize_symbol(symbol)
      if self.symbols.has_key(symbol):
         del self.symbols[symbol]
         self.trigger_callback('remove', tokens)
         return 1
      return 0

   def autocomplete(self, symbol):
      targets = self.symbols.keys()
      targets.sort()
      found_i = -1
      for i in range(len(targets)):
         if targets[i].startswith(symbol):
            found_i = i
            break
      if found_i != -1:
         found = targets[found_i]
         if found_i < len(targets) - 1:
            found_last = -1
            for i in range(found_i + 1, len(targets)):
               if targets[i].startswith(symbol):
                  found_last = i
               else:
                  break
            if found_last != -1:
               found_last = targets[found_last]
               found_it = 0
               for i in range(len(symbol), min(len(found), len(found_last))):
                  if found[i] != found_last[i]:
                     found_it = 1
                     break
               if found_it:
                  found = found[:i]
         return found
      return symbol

   def exists(self, symbol):
      return self.symbols.has_key(symbol)




# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2003 Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.core import *
from gamera import knn, gamera_xml, cluster, stats
import math
init_gamera()

def max_distance(glyphs, k):
   ud = k.unique_distances(glyphs, 0)
   ud.sort()
   return ud[-1]

def local_max_distance(glyphs, x, k):
   d = k.distance_from_images(glyphs, x)
   d.sort()
   return d[-1][0]

def get_graph_stats(glyphs, k):
   s = { }
   gc = knn.glyphs_by_category(glyphs)
   for x in gc.itervalues():
      if len(x) == 1:
         s[x[0].get_main_id()] = 1.0
         continue
      graph = cluster.make_spanning_tree(x, k)
      edges = []
      for edge in graph.get_edges():
         edges.append(edge.cost)
      id = x[0].get_main_id()
      s[id] = stats.mean(edges)
      #cluster.make_subtrees_stddev(graph, 1, 2, 0)
      #print graph.nedges, len(x)
   return s

def strip_small_categories(glyphs):
   gc = knn.glyphs_by_category(glyphs)
   new_glyphs = []
   for v in gc.itervalues():
      if len(v) > 3:
         new_glyphs.extend(v)
   return new_glyphs
         

def test():
   glyphs = gamera_xml.glyphs_from_xml(r"C:\Documents and Settings\Karl MacMillan\Desktop\test\prod.xml")

   glyphs = strip_small_categories(glyphs)
   from gamera.plugins import features
   k = knn.kNN()
   print k.features
   features.generate_features_list(glyphs, k.feature_functions)
   print "Getting gstats"

   graph_stats = get_graph_stats(glyphs, k)
   gstats = knn.get_glyphs_stats(glyphs)

   max_dist = max_distance(glyphs, k)
   print max_dist
   file = open("results.txt", "w")
   global_max = [[],[]]
   local_max = [[],[]]
   all = [[],[]]
   graph = [[],[]]
   gr_ccorrect = 0
   gr_icorrect = 0
   for x in glyphs:
      local_max_dist = local_max_distance(glyphs, x, k)
      ans = k.classify_with_images(glyphs, x, 1)
      file.write(ans[0][1] + ",")# + str(ans[0][0]) + ",")
      correct = 0
      if x.get_main_id() == ans[0][1]:
         file.write("1,")
         correct = 1
      else:
         file.write("0,")
      g = 1.0 - (ans[0][0] / max_dist)
      global_max[correct].append(g)
      file.write(str(g) + ",")

      l = 1.0 - (ans[0][0] / local_max_dist)
      local_max[correct].append(l)
      file.write(str(l) + ",")

      a = stats.samplestdev([ans[0][0],gstats[ans[0][1]][1]])
      all[correct].append(a)
      file.write(str(a) + ",")

      gr = stats.samplestdev([ans[0][0],graph_stats[ans[0][1]]])
      if (gr <= 1 and correct):
         gr_ccorrect += 1
      if (gr > 1 and not correct):
         gr_icorrect += 1
      graph[correct].append(gr)
      file.write(str(gr))

      file.write("\n")

   print "num correct: %d num incorrect: %d" % (len(global_max[1]), len(global_max[0]))
   print "confidence %f %f %f" % (((gr_ccorrect + gr_icorrect) / float(len(glyphs))),
                                  gr_ccorrect / float(len(glyphs) - len(global_max[0])),
                                  gr_icorrect / float(len(glyphs) - len(global_max[1])))

   cgm = -1
   igm = -1
   cgs = -1
   igs = -1
   if (len(global_max[0])):
      igm = stats.mean(global_max[0])
      igs = stats.samplestdev(global_max[0])
   if (len(global_max[1])):
      cgm = stats.mean(global_max[1])
      cgs = stats.samplestdev(global_max[1])

   clm = -1
   ilm = -1
   cls = -1
   ils = -1
   if (len(local_max[0])):
      ilm = stats.mean(local_max[0])
      ils = stats.samplestdev(local_max[0])
   if (len(local_max[1])):
      clm = stats.mean(local_max[1])
      cls = stats.samplestdev(local_max[1])

   cam = -1
   iam = -1
   cas = -1
   ias = -1
   if (len(all[0])):
      iam = stats.mean(all[0])
      ias = stats.samplestdev(all[0])
   if (len(all[1])):
      cam = stats.mean(all[1])
      cas = stats.samplestdev(all[1])

   cgraphm = -1
   igraphm = -1
   cgraphs = -1
   igraphs = -1
   if (len(graph[0])):
      igraphm = stats.mean(graph[0])
      igraphs = stats.samplestdev(graph[0])
   if (len(graph[1])):
      cgraphm = stats.mean(graph[1])
      cgraphs = stats.samplestdev(graph[1])

   print "global correct avg: %f stdev: %f incorrect avg: %f stddev: %f" % (cgm, cgs, igm, igs)
   print "local correct avg: %f stdev: %f incorrect avg: %f stddev: %f" % (clm, cls, ilm, ils)
   print "all correct avg: %f stdev: %f incorrect avg: %f stddev: %f" % (cam, cas, iam, ias)
   print "graph correct avg: %f stdev: %f incorrect avg: %f stddev: %f" % (cgraphm, cgraphs, igraphm, igraphs)

   def otsu_threshold(p):
      l = len(p)
      mu_T = 0.0
      for i in range(l):
         mu_T += i * p[i]

      sigma_T = 0.0
      for i in range(l):
         sigma_T += (i-mu_T)*(i-mu_T)*p[i]

      k_low = 0
      while (p[k_low] == 0) and (k_low < (l - 1)):
         k_low += 1
      k_high = l - 1
      while (p[k_high] == 0) and (k_high > 0):
         k_low += 1
         k_high -= 1

      criterion = 0.0
      thresh = 127

      omega_k = 0.0
      mu_k = 0.0
      k = k_low
      while k <= k_high:
         omega_k += p[k]
         mu_k += k*p[k]

         expr_1 = (mu_T*omega_k - mu_k)
         sigma_b_k = expr_1 * expr_1 / (omega_k*(1-omega_k))
         if (criterion < sigma_b_k/sigma_T):
            criterion = sigma_b_k/sigma_T
            thresh = k;
         k += 1
      return thresh

   graph_l = graph[0][:]
   graph_l.extend(graph[1])
   graph_l.sort()
   threshold = stats.mean(graph_l)
   print "threshold: " + str(threshold)
   num_wrong = 0
   for x in graph[0]:
      if x < threshold:
         num_wrong += 1
   print num_wrong, num_wrong / float(len(graph[0])) * 100

   num_wrong = 0
   for x in graph[1]:
      if x >= threshold:
         num_wrong += 1
   print num_wrong, num_wrong / float(len(graph[1])) * 100

   graph_l = all[0][:]
   graph_l.extend(all[1])
   graph_l.sort()
   threshold = stats.mean(graph_l)
   print "threshold: " + str(threshold)
   num_wrong = 0
   for x in graph[0]:
      if x < threshold:
         num_wrong += 1
   print num_wrong, num_wrong / float(len(graph[0])) * 100

   num_wrong = 0
   for x in graph[1]:
      if x >= threshold:
         num_wrong += 1
   print num_wrong, num_wrong / float(len(graph[1])) * 100

   graph_l = local_max[0][:]
   graph_l.extend(local_max[1])
   graph_l.sort()
   threshold = stats.mean(graph_l)
   print "threshold: " + str(threshold)
   num_wrong = 0
   for x in graph[0]:
      if x < threshold:
         num_wrong += 1
   print num_wrong, num_wrong / float(len(graph[0])) * 100

   num_wrong = 0
   for x in graph[1]:
      if x >= threshold:
         num_wrong += 1
   print num_wrong, num_wrong / float(len(graph[1])) * 100


#test()

   




#
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Most important problem with this module: Where do we find libpng?
#
# Linux:
#   We always assume libpng was installed in a default place (/usr or /usr/local)
#
# Mac OS-X:
#   We look for png.h in Fink (/sw/include/png.h) and then
#   /usr/include.  If it's not there, we use the local copy of libpng
#
# MS-Windows:
#   We always use the local copy of libpng

from gamera.plugin import *

class PNG_info(PluginFunction):
    """
    Returns an ``ImageInfo`` object describing a PNG file.

    *image_file_name*
      A PNG image filename
    """
    self_type = None
    args = Args([String("image_file_name")])
    return_type = ImageInfo("PNG_info")

class load_PNG(PluginFunction):
    """
    Loads a PNG format image file.

    *image_file_name*
      A PNG image filename

    *storage_format* (optional)
      specifies the compression type for the result:

      DENSE (0)
        no compression
      RLE (1)
        run-length encoding compression
    """
    self_type = None
    args = Args([FileOpen("image_file_name", "", "*.png"),
                 Choice("storage format", ["DENSE", "RLE"])])
    return_type = ImageType([ONEBIT, GREYSCALE, GREY16, RGB, FLOAT])
    def __call__(filename, compression = 0):
        from gamera.plugins import _png_support
        return _png_support.load_PNG(filename, compression)
    __call__ = staticmethod(__call__)
    exts = ['png']

class save_PNG(PluginFunction):
    """
    Saves the image to a PNG format file.
    """
    self_type = ImageType(ALL)
    args = Args([FileSave("image_file_name", "image.png", "*.png")])
    exts = ['png']

class PngSupportModule(PluginModule):
    import sys
    import os.path
    category = "File"
    cpp_headers = ["png_support.hpp"]
    internal_png_dir = "src/libpng-1.2.5/"
    internal_zlib_dir = "src/zlib-1.2.8/"
    if sys.platform == 'darwin':
        found_png = None
        for p in ["/sw", "/usr", "/usr/local"]:
            if os.path.exists(os.path.join(p, "include/png.h")):
                found_png = p
                break
        if found_png:
            cpp_include_dirs = [os.path.join(p, "include")]
            library_dirs = [os.path.join(p, "lib")]
            extra_libraries = ["png"]
        else:
            cpp_sources = [os.path.join(internal_png_dir, x) for x in
                           ['png.c', 'pngset.c', 'pngget.c', 'pngrutil.c',
                            'pngtrans.c', 'pngwutil.c', 'pngread.c', 'pngrio.c',
                            'pngwio.c', 'pngwrite.c', 'pngrtran.c', 'pngwtran.c',
                            'pngmem.c', 'pngerror.c', 'pngpread.c']]
            cpp_include_dirs = [internal_png_dir]
            # zlib, which apparently is included in OS-X 10.3 by default
            extra_libraries = ["z"]
    elif sys.platform in ('win32', 'cygwin'):
        cpp_sources = [os.path.join(internal_png_dir, x) for x in
                       ['png.c', 'pngset.c', 'pngget.c', 'pngrutil.c',
                        'pngtrans.c', 'pngwutil.c', 'pngread.c', 'pngrio.c',
                        'pngwio.c', 'pngwrite.c', 'pngrtran.c', 'pngwtran.c',
                        'pngmem.c', 'pngerror.c', 'pngpread.c',]]
        cpp_sources.extend([os.path.join(internal_zlib_dir, x) for x in
                            ['adler32.c','compress.c','crc32.c','deflate.c',
                             'gzclose.c','gzlib.c','gzread.c','gzwrite.c',
                             'infback.c','inffast.c','inflate.c','inftrees.c',
                             'trees.c','uncompr.c','zutil.c']])
        cpp_include_dirs = [internal_png_dir, "include/zlib-1.2.8"]
#        extra_libraries = ["z"]
    else:
        extra_libraries = ["png"]
    functions = [save_PNG, PNG_info, load_PNG]
    author = "Michael Droettboom and Albert Bzreckzo"
    url = "http://gamera.sourceforge.net/"
module = PngSupportModule()

PNG_info = PNG_info()
load_PNG = load_PNG()



# -*- mode: python; indent-tabs-mode: nil; tab-width: 4 -*-
# vim: set tabstop=4 shiftwidth=4 expandtab:

#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2007-2010 Christoph Dalitz and Uma Kompella
#               2014      Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
from gamera.args import NoneDefault
import _threshold

class threshold(PluginFunction):
    """
    Creates a binary image by splitting along a given global threshold value.

    Pixels that are greater than the given value become white.
    Pixels less than or equal to the given value become black.

    *storage_format* (optional)
      specifies the compression type for the result:

      DENSE (0)
        no compression
      RLE (1)
        run-length encoding compression.
    """
    self_type = ImageType([GREYSCALE, GREY16, FLOAT])
    args = Args([Int("threshold"), Choice("storage format", ['dense', 'rle'])])
    return_type = ImageType([ONEBIT], "output")
    doc_examples = [(GREYSCALE, 128)]
    def __call__(image, threshold, storage_format = 0):
        return _threshold.threshold(image, threshold, storage_format)
    __call__ = staticmethod(__call__)

class otsu_find_threshold(PluginFunction):
    """
    Finds a threshold point using the Otsu algorithm. Reference:

    N. Otsu: *A Threshold Selection Method from Grey-Level
    Histograms.* IEEE Transactions on Systems, Man, and Cybernetics
    (9), pp. 62-66 (1979)
    """
    self_type = ImageType([GREYSCALE])
    return_type = Int("threshold_point")
    doc_examples = [(GREYSCALE,)]

class otsu_threshold(PluginFunction):
    """
    Creates a binary image by splitting along a threshold value
    determined using the Otsu algorithm.

    Equivalent to ``image.threshold(image.otsu_find_threshold())``.

    *storage_format* (optional)
      specifies the compression type for the result:
      
      DENSE (0)
        no compression
      RLE (1)
        run-length encoding compression
    """
    self_type = ImageType([GREYSCALE])
    args = Args(Choice("storage format", ['dense', 'rle']))
    return_type = ImageType([ONEBIT], "output")
    doc_examples = [(GREYSCALE,)]
    def __call__(image, storage_format = 0):
        return _threshold.otsu_threshold(image, storage_format)
    __call__ = staticmethod(__call__)

class tsai_moment_preserving_find_threshold(PluginFunction):
    """
    Finds a threshold point using the Tsai Moment Preserving threshold
    algorithm. Reference:

    W.H. Tsai: *Moment-Preserving Thresholding: A New Approach.*
    Computer Vision Graphics and Image Processing (29), pp. 377-393
    (1985)
    """
    self_type = ImageType([GREYSCALE])
    return_type = Int("threshold_point")
    doc_examples = [(GREYSCALE,)]
    author = "Uma Kompella"

class tsai_moment_preserving_threshold(PluginFunction):
    """
    Creates a binary image by splitting along a threshold value
    determined using the Tsai Moment Preserving Threshold algorithm.

    Equivalent to
    ``image.threshold(image.tsai_moment_preserving_find_threshold())``.

    *storage_format* (optional)
      specifies the compression type for the result:

      DENSE (0)
        no compression
      RLE (1)
        run-length encoding compression
    """
    self_type = ImageType([GREYSCALE])
    args = Args(Choice("storage format", ['dense', 'rle']))
    return_type = ImageType([ONEBIT], "output")
    doc_examples = [(GREYSCALE,)]
    author = "Uma Kompella"
    def __call__(image, storage_format = 0):
        return _threshold.tsai_moment_preserving_threshold(image, storage_format)
    __call__ = staticmethod(__call__)




class abutaleb_threshold(PluginFunction):
    """
    Creates a binary image by using the Abutaleb locally-adaptive
    thresholding algorithm.

    *storage_format* (optional)
      specifies the compression type for the result:

      DENSE (0)
        no compression
      RLE (1)
        run-length encoding compression
    """
    self_type = ImageType([GREYSCALE])
    args = Args(Choice("storage format", ['dense', 'rle']))
    return_type = ImageType([ONEBIT], "output")
    doc_examples = [(GREYSCALE,)]
    def __call__(image, storage_format = 0):
        return _threshold.abutaleb_threshold(image, storage_format)
    __call__ = staticmethod(__call__)

class bernsen_threshold(PluginFunction):
    """
    Creates a binary image by using the Bernsen algorithm.

    Each point is thresholded by the mean between the maximum and minimum
    value in the surrounding region of size *region_size*. When the difference
    between maximum and minimum is below *contrast_limit* the pixel is set
    to black in case of *doubt_to_black* = ``True``, otherwise to white.

    Reference: J. Bernsen: *Dynamic thresholding of grey-level images.* 
    Proc. 8th International Conference on Pattern Recognition (ICPR8),
    pp. 1251-1255, 1986.

    *storage_format*
      specifies the compression type for the result:

      DENSE (0)
        no compression
      RLE (1)
        run-length encoding compression

    *region_size*
      The size of each region in which to calculate a threshold

    *contrast_limit*
      The minimum amount of contrast required to threshold.

    *doubt_to_black*
      When ``True``, 'doubtful' values are set to black, otherwise to white.
    """
    self_type = ImageType([GREYSCALE])
    args = Args([Choice("storage format", ['dense', 'rle']),
                 Int("region size", range=(1, 50), default=11),
                 Int("contrast limit", range=(0, 255), default=80),
                 Check("doubt_to_black", default=False)])
    return_type = ImageType([ONEBIT], "output")
    doc_examples = [(GREYSCALE,)]
    def __call__(image, storage_format = 0, region_size = 11,
                 contrast_limit = 80, doubt_to_black = False):
        return _threshold.bernsen_threshold(image, storage_format, region_size, contrast_limit, doubt_to_black)
    __call__ = staticmethod(__call__)

class djvu_threshold(PluginFunction):
    """
    Creates a binary image by using the DjVu thresholding algorithm.

    See Section 5.1 in:

      Bottou, L., P. Haffner, P. G. Howard, P. Simard, Y. Bengio and
      Y. LeCun.  1998.  High Quality Document Image Compression with
      DjVu.  AT&T Labs, Lincroft, NJ.

      http://research.microsoft.com/~patrice/PDF/jei.pdf

    This implementation features an additional extension to the
    algorithm described above.  Once the background and foreground
    colors are determined for each block, the image is thresholded by
    interpolating the foreground and background colors between the
    blocks.  This prevents "blockiness" along boundaries of strong
    color change.

    *smoothness*
      The amount of effect that parent blocks have on their children
      blocks.  Higher values will result in more smoothness between
      blocks.  Expressed as a percentage between 0.0 and 1.0.

    *max_block_size*
      The size of the largest block to determine a threshold.

    *min_block_size*
      The size of the smallest block to determine a threshold.

    *block_factor*
      The number of child blocks (in each direction) per parent block.
      For instance, a *block_factor* of 2 results in 4 children per
      parent.
    """
    self_type = ImageType([RGB])
    args = Args([Float("smoothness", default=0.2, range=(0.0, 1.0)),
                 Int("max_block_size", default=512),
                 Int("min_block_size", default=64),
                 Int("block_factor", default=2, range=(1, 8))])
    return_type = ImageType([ONEBIT], "output")
    def __call__(image, smoothness=0.2, max_block_size=512, min_block_size=64,
                 block_factor=2):
        return _threshold.djvu_threshold(image, smoothness, max_block_size,
                                         min_block_size, block_factor)
    __call__ = staticmethod(__call__)
    doc_examples = [(RGB, 0.5, 512, 64, 2)]

class soft_threshold(PluginFunction):
    """
    Does a greyscale transformation that \"smears out\" the threshold *t* by a
    choosable amount *sigma*. This has the effect of a \"soft\" thresholding.

    Each grey value *x* is transformed to *F(x,t,sigma)*, where *F*
    is the CDF probability distribution with mean *t* and variance
    *sigma^2*. The parameter *dist* determines the type of probability
    distribution: 0 = logistic, 1 = normal (gaussian), 2 = uniform.

    As the choice *sigma* = 0 is useless (it is the same as normal
    thresholding), this special value is reserved for an automatic selection
    of *sigma* with soft_threshold_find_sigma__.

.. __: #soft-threshold-find-sigma

    When *t* is not given, it is automatically computed with
    otsu_find_threshold__.

.. __: #otsu-find-threshold

    Reference: C. Dalitz: `\"Soft Thresholding for Visual Image Enhancement.\"`__
    Technischer Bericht Nr. 2014-01, Hochschule Niederrhein,
    Fachbereich Elektrotechnik und Informatik, 2014

.. __: http://www.hsnr.de/fileadmin/dateien/fb03/gruppen/technische-berichte/fb03-tb-2014-01.pdf
    """
    self_type = ImageType([GREYSCALE])
    args = Args([Int("t", default=NoneDefault), Float("sigma", default=0.0), Choice("dist", ["logistic","normal","uniform"], default=0)])
    return_type = ImageType([GREYSCALE], "output")
    author = "Christoph Dalitz"
    def __call__(image, t=None, sigma=0.0, dist=0):
        if t is None:
            t = image.otsu_find_threshold()
        return _threshold.soft_threshold(image, t, sigma, dist)
    __call__ = staticmethod(__call__)
    doc_examples = [(GREYSCALE, 128, 25)]

class soft_threshold_find_sigma(PluginFunction):
    """
    For the CDF probability distribution given by *dist*
    (0 = logistic, 1 = normal (gaussian), 2 = uniform), sigma is
    determined such that *F(m,t,sigma)* = 0.99, where *m* is the mean grey
    value of all pixels with a grey value greater than *t*.

    Reference: C. Dalitz: `\"Soft Thresholding for Visual Image Enhancement.\"`__
    Technischer Bericht Nr. 2014-01, Hochschule Niederrhein,
    Fachbereich Elektrotechnik und Informatik, 2014

.. __: http://www.hsnr.de/fileadmin/dateien/fb03/gruppen/technische-berichte/fb03-tb-2014-01.pdf
    """
    self_type = ImageType([GREYSCALE])
    args = Args([Int("t", default=NoneDefault), Choice("dist", ["logistic","normal","uniform"], default=0)])
    return_type = Float("sigma")
    author = "Christoph Dalitz"
    def __call__(image, t, dist=0):
        return _threshold.soft_threshold_find_sigma(image, t, dist)
    __call__ = staticmethod(__call__)


class ThresholdModule(PluginModule):
    """
    This module provides functions that convert images between different
    pixel types.
    """
    category = "Binarization"
    cpp_headers = ["threshold.hpp"]
    functions = [threshold, otsu_find_threshold, otsu_threshold,
                 tsai_moment_preserving_find_threshold,
                 tsai_moment_preserving_threshold, abutaleb_threshold,
                 bernsen_threshold, djvu_threshold,
                 soft_threshold, soft_threshold_find_sigma]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"

module = ThresholdModule()



#
# Copyright (C) 2005 Alex Cobb
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""Data transfer between Gamera and Numeric.
"""

from gamera.plugin import *
from gamera.util import warn_deprecated
from gamera import config

try:
    import numpy.oldnumeric as n
except ImportError:
    try:
        verbose = config.get("verbosity_level")
    except Exception:
        verbose = 0
    if verbose:
        print ('Info: numpy.oldumeric could not be imported.')
else:
    _typecodes = {RGB       : n.UInt8,
                  GREYSCALE : n.UInt8,
                  GREY16    : n.UInt32,
                  ONEBIT    : n.UInt16,
                  FLOAT     : n.Float64,
                  COMPLEX   : n.Complex64 }
    _inverse_typecodes = { n.UInt8     : GREYSCALE,
                           n.UInt32    : GREY16,
                           n.UInt16    : ONEBIT,
                           n.Float64   : FLOAT,
                           n.Complex64 : COMPLEX } 
        
    class from_numeric(PluginFunction):
        """
        Instantiates a Gamera image from a Numeric multi-dimensional
        array *array*.
            
        The array must be one of the following types and will map to
        the corresponding Gamera image type:

        +------------+------------------+
        | Gamera     | Numeric          |
        | type       | type             |
        +============+==================+
        | RGB        | UInt8 (on 3      |
        |            | planes)          |
        +------------+------------------+
        | GREYSCALE  | UInt8            |
        +------------+------------------+
        | GREY16     | UInt32           |
        +------------+------------------+
        | ONEBIT     | UInt16           |
        +------------+------------------+
        | FLOAT      | Float64          |
        +------------+------------------+
        | COMPLEX    | Complex64        |
        +------------+------------------+

        Requires two copying operations;  may fail for very large images.

        To use this function, which is not a method on images, do the
        following:

        .. code:: Python
        
          from gamera.plugins import numeric_io
          image = numeric_io.from_numeric(array)
        """
        self_type = None
        args = Args([Class("array")])
        return_type = ImageType(ALL)
        pure_python = True
        def __call__(array, offset=(0, 0)):
            from gamera.plugins import _string_io
            from gamera.core import Dim
            pixel_type = from_numeric._check_input(array)
            return _string_io._from_raw_string(
                offset,
                Dim(array.shape[1], array.shape[0]),
                pixel_type, DENSE,
                array.tostring())
        __call__ = staticmethod(__call__)

        def _check_input(array):

            shape = array.shape
            typecode = array.dtype.char
            if len(shape) == 3 and shape[2] == 3 and typecode == n.UInt8:
                return RGB
            elif len(shape) == 2:
                if _inverse_typecodes.has_key(typecode):
                    return _inverse_typecodes[typecode]
            raise ValueError('Array is not one of the acceptable types (UInt8 * 3, UInt8, UInt16, UInt32, Float64, Complex64)')
        _check_input = staticmethod(_check_input)

    class to_numeric(PluginFunction):
        """
        Returns an ``Numeric`` array containing a copy of the image's data.

        The array will be one of the following types corresponding to
        each of the Gamera image types:

        +------------+-----------------+
        | Gamera     | Numeric         |
        | type       | type            |
        +============+=================+
        | RGB        | UInt8 (on 3     |
        |            | planes)         |
        +------------+-----------------+
        | GREYSCALE  | UInt8           |
        +------------+-----------------+
        | GREY16     | UInt32          |
        +------------+-----------------+
        | ONEBIT     | UInt16          |
        +------------+-----------------+
        | FLOAT      | Float64         |
        +------------+-----------------+
        | COMPLEX    | Complex64       |
        +------------+-----------------+

        Requires *three* copies, and may fail for very large images.
        """
        self_type = ImageType(ALL)
        return_type = Class("array")
        pure_python = True
        def __call__(image):
            from gamera.plugins import _string_io
            pixel_type = image.data.pixel_type
            shape = (image.nrows, image.ncols)
            typecode = _typecodes[pixel_type]
            if pixel_type == RGB:
                shape += (3,)
            array = n.fromstring(_string_io._to_raw_string(image), typecode)
            return n.resize(array, shape)
        __call__ = staticmethod(__call__)

        def __doc_example1__(images):
            image = images[RGB]
            array = image.to_numeric()
            image0 = from_numeric(array)
            return [image, image0]
        def __doc_example2__(images):
            image = images[GREYSCALE]
            array = image.to_numeric()
            image0 = from_numeric(array)
            return [image, image0]
        def __doc_example4__(images):
            image = images[ONEBIT]
            array = image.to_numeric()
            image0 = from_numeric(array)
            return [image, image0]
        def __doc_example5__(images):
            image = images[FLOAT]
            array = image.to_numeric()
            image0 = from_numeric(array)
            return [image, image0]
        def __doc_example6__(images):
            image = images[COMPLEX]
            array = image.to_numeric()
            image0 = from_numeric(array)
            return [image, image0]
        doc_examples = [__doc_example1__,
                        __doc_example2__,
                        __doc_example4__,
                        __doc_example5__,
                        __doc_example6__]

    class NumericModule(PluginModule):
        category = None #"ExternalLibraries/Numeric"
        author = "Alex Cobb"
        functions = [from_numeric, to_numeric]
        pure_python = True
        url = ('http://www.oeb.harvard.edu/faculty/holbrook/'
               'people/alex/Website/alex.htm')
    module = NumericModule()

    from_numeric = from_numeric()



from gamera.core import *

from gamera.core import *
init_gamera()

def test_nested_list():
   def _test_nested_list(name, type, storage = DENSE):
      image = load_image("data/%s_generic.tiff" % name, storage)
      nested = image.to_nested_list()
      image2 = nested_list_to_image(nested, type)
      assert image._to_raw_string() == image2._to_raw_string()

   # TODO: Add Grey16
   for name, type in [("OneBit", ONEBIT), ("GreyScale", GREYSCALE), ("RGB", RGB)]:
      _test_nested_list(name, type)
   _test_nested_list("OneBit", ONEBIT, RLE)



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""The relational module contains plugins for computing the relationships
between glyphs."""

from gamera.plugin import * 

class bounding_box_grouping_function(PluginFunction):
    """
    Given two rectangles *a*, *b*, and a given *threshold* distance
    (in pixels), returns ``True`` if the two rectangles are closer
    than *threshold*.
    """
    self_type = None
    args = Args([Rect("a"), Rect("b"), Int("threshold")])
    return_type = Check("connected")

class shaped_grouping_function(PluginFunction):
    """
    Given two connected components *a*, *b*, and a given *threshold*
    distance (in pixels), returns ``True`` if any pixel in *a* are
    closer than *threshold* to any pixel in *b*.
    """
    self_type = None
    args = Args([ImageType(ONEBIT, "a"), ImageType(ONEBIT, "b"), Int("threshold")])
    return_type = Check("connected")

class polar_distance(PluginFunction):
    """
    Returns a tuple containing the normalized distance, polar
    direction, and non-normalized polar distance to another glyph
    (based on center of bounding boxes).
    """
    self_type = ImageType(ALL)
    return_type = FloatVector("polar")
    args = Args([ImageType(ALL, "other")])

class polar_match(PluginFunction):
    self_type = None
    return_type = Int("check")
    args = Args([Float('r1'), Float('q1'), Float('r2'), Float('q2')])

class least_squares_fit(PluginFunction):
    """
    Performs a least squares fit on a given list of points.

    The result is a tuple of the form (*m*, *b*, *q*) where *m* is the
    slope of the line, *b* is the *y*-offset, and *q* is the gamma fit
    of the line to the points. (This assumes the same statistical
    significance for all points).
    
    See Numerical Recipes in C, section 15.2__ for more information.

    .. __: http://www.library.cornell.edu/nr/bookcpdf/c15-2.pdf
    """
    self_type = None
    return_type = Class("a_b_q")
    args = Args([PointVector("points")])

class least_squares_fit_xy(PluginFunction):
    """
    Identical to *least_squares_fit* for line angles below 45 degrees.
    For lines with a more vertical slope a least square fit of *x = my
    + b* is done instead.

    The result is a tuple of the form (*m*, *b*, *q*, *x_of_y*) where
    *m, b, q* are the same as in *least_squares_fit*, but the integer
    value *x_of_y* determines the actual meaning of the parameters *m*
    and *b*:

    When *x_of_y* is zero, *y = mx + b*. Otherwise *x = my + b*.
    """
    self_type = None
    return_type = Class("a_b_q_xofy")
    args = Args([PointVector("points")])
    author = "Christoph Dalitz"

class edit_distance(PluginFunction):
    """
    Computes the edit distance (also known as *Levenshtein distance*) between
    two strings.

    This counts the number of character substitutions, additions and deletions
    necessary to transform one string into another. This plugin is a 
    straightforward implementation of the classic algorithm by Wagner 
    and Fischer, which has runtime complexity *O(m*n)*, where *m* and *n* are
    the lengths of the two strings.

    See R.A. Wagner, M.J. Fischer: *The String-to-String Correction Problem.*
    Journal of the ACM 21, pp. 168-173, 1974.
    """
    self_type = None
    args = Args([String("s1"), String("s2")])
    return_type = Int("distance")
    author = "Christoph Dalitz"

class RelationalModule(PluginModule):
    cpp_headers = ["structural.hpp"]
    category = "Relational"
    functions = [polar_distance, polar_match,
                 bounding_box_grouping_function,
                 shaped_grouping_function,
                 least_squares_fit, least_squares_fit_xy,
                 edit_distance]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"

module = RelationalModule()

bounding_box_grouping_function = bounding_box_grouping_function()
shaped_grouping_function = shaped_grouping_function()
least_squares_fit = least_squares_fit()
least_squares_fit_xy = least_squares_fit_xy()
edit_distance = edit_distance()
